/* SPDX-License-Identifier: LGPL-2.1+ */
/******************************************************************************
 * Copyright (c) Huawei Technologies Co., Ltd. 2020. Allrights reserved
 * Description: isulad utils
 * Author: lifeng
 * Create: 2020-04-11
******************************************************************************/

#ifndef _GNU_SOURCE
#define _GNU_SOURCE 1
#endif

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <stdio_ext.h>

#include "isulad_utils.h"
#include "log.h"
#include "path.h"
#include "file_utils.h"

lxc_log_define(isulad_utils, lxc);

void *lxc_common_calloc_s(size_t size)
{
	if (size == 0 || size > SIZE_MAX) {
		return NULL;
	}

	return calloc((size_t)1, size);
}

int lxc_mem_realloc(void **newptr, size_t newsize, void *oldptr, size_t oldsize)
{
	void *tmp = NULL;

	if (newptr == NULL || newsize == 0) {
		return -1;
	}

	tmp = lxc_common_calloc_s(newsize);
	if (tmp == NULL) {
		ERROR("Failed to malloc memory");
		return -1;
	}

	if (oldptr != NULL) {
		memcpy(tmp, oldptr, (newsize < oldsize) ? newsize : oldsize);

		memset(oldptr, 0, oldsize);

		free(oldptr);
	}

	*newptr = tmp;
	return 0;
}

char *safe_strdup(const char *src)
{
	char *dst = NULL;

	if (src == NULL) {
		return NULL;
	}

	dst = strdup(src);
	if (dst == NULL) {
		abort();
	}

	return dst;
}

int lxc_open(const char *filename, int flags, mode_t mode)
{
	char rpath[PATH_MAX] = {0x00};

	if (filename == NULL) {
		return -1;
	}

	if (cleanpath(filename, rpath, sizeof(rpath)) == NULL) {
		return -1;
	}
	if (mode) {
		return open(rpath, (int)((unsigned int)flags | O_CLOEXEC), mode);
	} else {
		return open(rpath, (int)((unsigned int)flags | O_CLOEXEC));
	}
}

FILE *lxc_fopen(const char *filename, const char *mode)
{
	char rpath[PATH_MAX] = {0x00};

	if (filename == NULL || mode == NULL || strlen(mode) == 0) {
		return NULL;
	}

	if (cleanpath(filename, rpath, sizeof(rpath)) == NULL) {
		return NULL;
	}

	return fopen_cloexec(rpath, mode);
}

/* isulad: write error message */
void lxc_write_error_message(int errfd, const char *format, ...)
{
	int ret;
	char errbuf[BUFSIZ + 1] = {0};
	ssize_t sret;
	va_list argp;

	if (errfd <= 0)
		return;

	va_start(argp, format);
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wformat-nonliteral"
	ret = vsnprintf(errbuf, BUFSIZ, format, argp);
#pragma GCC diagnostic pop
	va_end(argp);
	if (ret < 0 || ret >= BUFSIZ)
		SYSERROR("Failed to call vsnprintf");
	sret = write(errfd, errbuf, strlen(errbuf));
	if (sret < 0)
		SYSERROR("Write errbuf failed");
}

/* isulad: read file to buffer */
int lxc_file2str(const char *filename, char ret[], int cap)
{
    int fd, num_read;

    if (filename == NULL || ret == NULL || cap == 0) {
        return -1;
    }

    if ((fd = lxc_open(filename, O_RDONLY | O_CLOEXEC, 0)) == -1)
        return -1;
    if ((num_read = read(fd, ret, cap - 1)) <= 0)
        num_read = -1;
    else
        ret[num_read] = 0;
    close(fd);

    return num_read;
}

/* isuald: lxc_stat2proc() makes sure it can handle arbitrary executable file basenames
 * for `cmd', i.e. those with embedded whitespace or embedded ')'s.
 * Such names confuse %s (see scanf(3)), so the string is split and %39c
 * is used instead. (except for embedded ')' "(%[^)]c)" would work.
 */
static proc_t *lxc_stat2proc(const char *S)
{
    int num;
    proc_t *P = NULL;
    char *tmp = NULL;

    if (!S)
        return NULL;

    tmp = strrchr(S, ')');  /* split into "PID (cmd" and "<rest>" */
    if (!tmp)
        return NULL;
    *tmp = '\0';            /* replace trailing ')' with NUL */

    P = malloc(sizeof(proc_t));
    if (P == NULL)
        return NULL;
    (void)memset(P, 0x00, sizeof(proc_t));

    /* parse these two strings separately, skipping the leading "(". */
    /* https://www.openwall.com/lists/musl/2013/11/15/5: musl's sscanf("%15c",cmd) 
       requires exactly 15 characters; anything shorter is a matching failure. */
#ifdef __MUSL__
    num = sscanf(S, "%d (%15s", &P->pid, P->cmd);   /* comm[16] in kernel */
#else
    num = sscanf(S, "%d (%15c", &P->pid, P->cmd);   /* comm[16] in kernel */
#endif
    if (num != 2) {
        ERROR("Call sscanf error: %s", errno ? strerror(errno) : "");
        free(P);
        return NULL;
    }
    num = sscanf(tmp + 2,           /* skip space after ')' too */
                 "%c "
                 "%d %d %d %d %d "
                 "%lu %lu %lu %lu %lu "
                 "%Lu %Lu %Lu %Lu "  /* utime stime cutime cstime */
                 "%ld %ld %ld %ld "
                 "%Lu "  /* start_time */
                 "%lu "
                 "%ld "
                 "%lu %lu %lu %lu %lu %lu "
                 "%*s %*s %*s %*s " /* discard, no RT signals & Linux 2.1 used hex */
                 "%lu %lu %lu "
                 "%d %d "
                 "%lu %lu",
                 &P->state,
                 &P->ppid, &P->pgrp, &P->session, &P->tty, &P->tpgid,
                 &P->flags, &P->min_flt, &P->cmin_flt, &P->maj_flt, &P->cmaj_flt,
                 &P->utime, &P->stime, &P->cutime, &P->cstime,
                 &P->priority, &P->nice, &P->timeout, &P->it_real_value,
                 &P->start_time,
                 &P->vsize,
                 &P->rss,
                 &P->rss_rlim, &P->start_code, &P->end_code, &P->start_stack, &P->kstk_esp,
                 &P->kstk_eip,
                 &P->wchan, &P->nswap, &P->cnswap,
                 &P->exit_signal, &P->processor,  /* 2.2.1 ends with "exit_signal" */
                 &P->rtprio, &P->sched  /* both added to 2.5.18 */
                );
    if (num != 35) {
        ERROR("Call sscanf error: %s", errno ? strerror(errno) : "");
        free(P);
        return NULL;
    }
    if (P->tty == 0)
        P->tty = -1;  /* the old notty val, update elsewhere bef. moving to 0 */
    return P;
}

/* isulad: get starttime of process pid */
unsigned long long lxc_get_process_startat(pid_t pid)
{
    int sret = 0;
    unsigned long long startat = 0;
    proc_t *pid_info = NULL;
    char filename[PATH_MAX] = {0};
    char sbuf[1024] = {0}; /* bufs for stat */

    sret = snprintf(filename, sizeof(filename), "/proc/%d/stat", pid);
    if (sret < 0 || sret >= sizeof(filename)) {
        ERROR("Failed to sprintf filename");
        goto out;
    }

    if ((lxc_file2str(filename, sbuf, sizeof(sbuf))) == -1) {
        SYSERROR("Failed to read pidfile %s", filename);
        goto out;
    }

    pid_info = lxc_stat2proc(sbuf);
    if (!pid_info) {
        ERROR("Failed to get proc stat info");
        goto out;
    }

    startat = pid_info->start_time;
out:
    free(pid_info);
    return startat;
}

// isulad: set env home in container
int lxc_setup_env_home(uid_t uid)
{
#define __PASSWD_FILE__ "/etc/passwd"
	char *homedir = "/"; // default home dir is /
	FILE *stream = NULL;
	struct passwd pw, *pwbufp = NULL;
	char buf[BUFSIZ];
	const char *curr_home = NULL;

	curr_home = getenv("HOME");
	// if user set or image set, just use it.
	if (curr_home != NULL && strcmp(curr_home, "") != 0) {
		return 0;
	}

	stream = fopen_cloexec(__PASSWD_FILE__, "r");
	if (stream == NULL) {
		SYSWARN("Failed to open %s", __PASSWD_FILE__);
		goto set_env;
	}

#if IS_BIONIC
	while (util_getpwent_r(stream, &pw, buf, sizeof(buf), &pwbufp) == 0 && pwbufp != NULL) {
#else
	while (fgetpwent_r(stream, &pw, buf, sizeof(buf), &pwbufp) == 0 && pwbufp != NULL) {
#endif
		if (pwbufp->pw_uid == uid) {
			homedir = pwbufp->pw_dir;
			goto set_env;
		}
	}
	WARN("User invalid, can not find user '%u'", uid);

set_env:
	if (stream)
		fclose(stream);

	// if we didn't configure HOME, set it based on uid;
	// override it if reach here.
	if (setenv("HOME", homedir, 1) < 0) {
		SYSERROR("Unable to set env 'HOME'");
		return -1;
	}

	NOTICE("Setted env 'HOME' to %s", homedir);
	return 0;
}

bool lxc_process_alive(pid_t pid, unsigned long long start_time)
{
	int sret = 0;
	bool alive = true;
	proc_t *pid_info = NULL;
	char filename[PATH_MAX] = {0};
	char sbuf[1024] = {0}; /* bufs for stat */

	sret = kill(pid, 0);
	if (sret < 0 && errno == ESRCH)
		return false;

	sret = snprintf(filename, sizeof(filename), "/proc/%d/stat", pid);
	if (sret < 0 || sret >= sizeof(filename)) {
		ERROR("Failed to sprintf filename");
		goto out;
	}

	if ((lxc_file2str(filename, sbuf, sizeof(sbuf))) == -1) {
		ERROR("Failed to read pidfile %s", filename);
		alive = false;
		goto out;
	}

	pid_info = lxc_stat2proc(sbuf);
	if (!pid_info) {
		ERROR("Failed to get proc stat info");
		alive = false;
		goto out;
	}

	if (start_time != pid_info->start_time)
		alive = false;
out:
	free(pid_info);
	return alive;
}

bool is_non_negative_num(const char *s)
{
	if (!s || !strcmp(s, ""))
		return false;
	while(*s  != '\0') {
		if(!isdigit(*s))
			return false;
		++s;
	}
	return true;
}

static int hold_int(const char delim, bool required, char **src, unsigned int *dst)
{
    unsigned long long int res = 0;
    char *err_str = NULL;

    // ensure *src not a empty string
    if (**src == '\0') {
        ERROR("Empty subject on given entrie is not allowed.");
        return -1;
    }

    errno = 0;
    // covert string to long long
    res = strtoull(*src, &err_str, 0);
    if (errno != 0 && errno != ERANGE) {
        ERROR("Parse int from string failed.");
        return -1;
    }

    // **src is not a digit
    if (err_str == *src) {
        if (!required) {
            ERROR("Integer part is missing.");
            return -1;
        }
        // if required, just set 0
        *dst = 0;
    } else {
        if (sizeof(void *) > 4 && res > UINT_MAX) { // make sure 64-bit platform behave same as 32-bit
            res = UINT_MAX;
        }
        res = res & UINT_MAX;
        *dst = (uint32_t)res;
    }

    // normal case
    if (*err_str == delim) {
        err_str++;
    } else if (*err_str != '\0') {
        ERROR("Invalid digit string.");
        return -1;
    }

    *src = err_str; // update src to next valid context in line.
    return 0;
}

static void hold_string(const char delim, char **src, char **dst)
{
    for (*dst = *src; **src != delim; ++(*src)) {
        if (**src == '\0') {
            break;
        }
    }

    if (**src == delim) {
        **src = '\0';
        ++(*src);
    }
}

static int parse_line_pw(const char delim, char *line, struct passwd *result)
{
    int ret = 0;
    bool required = false;
    char *walker = NULL;

    walker = strpbrk(line, "\n");
    if (walker != NULL) {
        // clear newline char
        *walker = '\0';
    }

    hold_string(delim, &line, &result->pw_name);

    required = (result->pw_name[0] == '+' || result->pw_name[0] == '-') ? true : false;

    hold_string(delim, &line, &result->pw_passwd);

    ret = hold_int(delim, required, &line, &result->pw_uid);
    if (ret != 0) {
        // a legitimate line must have uid
        ERROR("Parse uid error.");
        return ret;
    }

    ret = hold_int(delim, required, &line, &result->pw_gid);
    if (ret != 0) {
        // it's ok to not provide gid
        ERROR("Parse gid error.");
        return ret;
    }

    hold_string(delim, &line, &result->pw_gecos);

    hold_string(delim, &line, &result->pw_dir);

    result->pw_shell = line;
    return 0;
}

static char *util_left_trim_space(char *str)
{
    char *begin = str;
    char *tmp = str;
    while (isspace(*begin)) {
        begin++;
    }
    while ((*tmp++ = *begin++)) {
    }
    return str;
}

int util_getpwent_r(FILE *stream, struct passwd *resbuf, char *buffer, size_t buflen, struct passwd **result)
{
    const char delim = ':';
    char *buff_end = NULL;
    char *walker = NULL;
    bool got = false;
    int ret = 0;

    if (stream == NULL || resbuf == NULL || buffer == NULL || result == NULL) {
        ERROR("Password obj, params is NULL.");
        return -1;
    }

    if (buflen <= 1) {
        ERROR("Inadequate buffer length was given.");
        return -1;
    }

    buff_end = buffer + buflen - 1;
    flockfile(stream);

    while (1) {
        *buff_end = '\xff';
        walker = fgets_unlocked(buffer, buflen, stream);
        // if get NULL string
        if (walker == NULL) {
            *result = NULL;
            // reach end of file, return error
            if (feof(stream)) {
                ret = ENOENT;
                goto out;
            }
            // overflow buffer
            ret = ERANGE;
            goto out;
        }
        // just overflow last char in buffer
        if (*buff_end != '\xff') {
            *result = NULL;
            ret = ERANGE;
            goto out;
        }

        (void)util_left_trim_space(buffer);
        // skip comment line and empty line
        if (walker[0] == '#' || walker[0] == '\0') {
            continue;
        }

        if (parse_line_pw(delim, walker, resbuf) == 0) {
            got = true;
            break;
        }
    }
    if (!got) {
        *result = NULL;
        ret = ERANGE;
        goto out;
    }

    *result = resbuf;
    ret = 0;
out:
    funlockfile(stream);
    return ret;
}
